Program DeZip;

{   DeZip v1.5 (C) Copyright 1989 by R. P. Byrne                              }
{                                                                             }
{   This is a "bare-bones" program to extract files from ZIP archives.        }
{   By "bare-bones", I mean that there is no facility included to do anything }
{   but extraction (ie. no echo to console, no send to printer, etc.).        }
{   If relative pathnames are stored in the Zip file, make sure all of the    }
{   required directories exist on your system before attempting an            }
{   extraction.                                                               }

{$B-,R-,S-}

Uses
   Dos,
   Crt,
   ZipHead,
   StrProcs;

Const
   COPYRIGHT = 'DeZip (C) Copyright 1989 by R. P. Byrne';
   VERSION   = 'Version 1.5 - Compiled on March 11, 1989';

{ Stuff needed generically by all uncompression methods }

Type
   BufPtr         =  ^BufType;
   BufType        =  Array[1..BUFSIZE] of Byte;

Var
   InFileSpecs    :  Array[1..MAXNAMES] of String;   { Input file specifications }
   MaxSpecs       :  Word;           { Total number of entries in InFileSpecs array }
   OutPath        :  String;         { Output path specification }
   ZipName        :  String;         { Name of Zip file to be processed }
   ZipFile        :  File;           { Zip file variable }
   EndFile        :  Boolean;        { End of file indicator for ZipFile }
   ZipBuf         :  BufPtr;         { Input buffer for ZipFile }
   ZipPtr         :  Word;           { Index for ZipFile input buffer }
   ZipCount       :  Word;           { Count of bytes in ZipFile input buffer }

   ExtFile        :  File;           { Output file variable }
   ExtBuf         :  BufPtr;         { Output buffer for ExtFile }
   ExtPtr         :  Word;           { Index for ExtFile output buffer }
   ExtCount       :  LongInt;        { Count of characters written to output }
   TenPercent     :  LongInt;

   LocalHdr       :  Local_File_Header_Type;  { Storage for a local file hdr }
   Hdr_FileName   :  String;
   Hdr_ExtraField :  String;
   Hdr_Comment    :  String;

   Bytes_To_Go    :  LongInt;        { Bytes left to process in compressed file }


{ Stuff needed for unSHRINKing }

Type
   LZW_Table_Rec  =  Record
                        Prefix      :  Integer;
                        Suffix      :  Byte;
                        ChildCount  :  Word;  { If ChildCount = 0 then leaf node }
                     end;
   LZW_Table_Ptr  =  ^LZW_Table_Type;
   LZW_Table_Type =  Array[0..TABLESIZE] of LZW_Table_Rec;

   FreeListPtr    =  ^FreeListArray;
   FreeListArray  =  Array[FIRSTENTRY..TABLESIZE] of Word;

   StackPtr       =  ^StackType;
   StackType      =  Array[0..TABLESIZE] of Word;

Var
   LZW_Table  :  LZW_Table_Ptr; { Code table for LZW decoding                }
   FreeList   :  FreeListPtr;   { List of free table entries                 }
   NextFree   :  Word;          { Index for free list array                  }
                                {   FreeList^[NextFree] always contains the  }
                                {   index of the next available entry in     }
                                {   the LZW Prefix:Suffix table (LZW_Table^) }
   LZW_Stack  :  StackPtr;      { A stack used to build decoded strings      }
   StackIdx   :  Word;          { Stack array index variable                 }
                                {   StackIdx always points to the next       }
                                {   available entry in the stack             }
   SaveByte   :  Byte;          { Our input code buffer - 1 byte long        }
   BitsLeft   :  Byte;          { Unprocessed bits in the input code buffer  }
   FirstCh    :  Boolean;       { Flag indicating first char being processed }


{ Stuff needed for unREDUCEing }

Const
   RedMask : Array[1..4] of Byte = ($7F, $3F, $1F, $0F);

Type
   FollowerSet    =  Record
                        SetSize  :  Word;
                        FSet     :  Array[0..31] of Byte;
                     end;
   FollowerPtr    =  ^FollowerArray;
   FollowerArray  =  Array[0..255] of FollowerSet;

   StreamPtr      =  ^StreamArray;
   StreamArray    =  Array[0..4095] of Byte;

Var
   Followers   :  FollowerPtr;
   Stream      :  StreamPtr;     { The output stream }
   StreamIdx   :  Word;          { Always points to next pos. to be filled }
   State       :  Byte;
   Len         :  Word;
   V           :  Byte;

{ --------------------------------------------------------------------------- }

Procedure Abort(Msg : String);
Begin
   Writeln;
   Writeln(Msg);
   Writeln('Returning to DOS');
   Writeln;
   Halt;
end {Abort};

{ --------------------------------------------------------------------------- }

Procedure Syntax;
Begin
   Writeln('Usage:  DeZip ZipFileName [OutPathSpec] [FileSpec [...]]');
   Writeln;
   Writeln('Optional file specifications may contain DOS ');
   Writeln('wildcard characters.');
   Writeln;
   Writeln('If no filespecs are entered, *.* is assumed.');
   Writeln;
   Halt;
End;

{ --------------------------------------------------------------------------- }

Function IO_Test : Boolean;
Var
   ErrorCode : Word;
   Ok        : Boolean;
Begin
   Ok := TRUE;
   ErrorCode := IOResult;
   If ErrorCode <> 0 then begin
      Ok := FALSE;
      Case ErrorCode of
           2 : Writeln('File Not Found');
           3 : Writeln('Path Not Found');
           5 : Writeln('Access Denied');
         101 : Writeln('Disk Full');
        else   Writeln('I/O Error #', ErrorCode);
      end {Case};
   end {if};
   IO_Test := Ok;
end {IO_Test};

{ --------------------------------------------------------------------------- }

Procedure Load_Parms;
Var
   I      : Word;
   Name   : String;
   DosDTA : SearchRec;
Begin
   I := ParamCount;
   If I < 1 then
      Syntax;

   ZipName := ParamStr(1);
   For I := 1 to Length(ZipName) do
      ZipName[I] := UpCase(ZipName[I]);
   If Pos('.', ZipName) = 0 then
      ZipName := ZipName + '.ZIP';

   MaxSpecs := 0;
   OutPath := '';
   I := 1;
   While I < ParamCount do begin
      Inc(I);
      Name := ParamStr(I);
      If Name[length(Name)] = '\' then
         Delete(Name, Length(Name), 1);
      FindFirst(Name, DIRECTORY, DosDTA);   { outpath spec? }
      If DosError = 0 then begin
         If (DosDTA.Attr AND DIRECTORY) <> 0 then begin   { yup }
            OutPath := Name;
            If OutPath[Length(OutPath)] <> '\' then
               OutPath := OutPath + '\';
         end {then}
         else begin
            If MaxSpecs < MAXNAMES then begin
               Inc(MaxSpecs);
               InFileSpecs[MaxSpecs] := Name;
            end {if};
         end {if};
      end {then}
      else begin
         If MaxSpecs < MAXNAMES then begin
            Inc(MaxSpecs);
            InFileSpecs[MaxSpecs] := Name;
         end {if};
      end {if}
   end {while};

   If MaxSpecs = 0 then begin
      MaxSpecs := 1;
      InFileSpecs[1] := '*.*';
   end {if};
end {Load_Parms};

{ --------------------------------------------------------------------------- }

Procedure Initialize;
Const
   Msg = 'Not enough memory to allocate I/O buffers!';
Begin
   If SizeOf(ZipBuf^) > MaxAvail then
      Abort(Msg);
   New(ZipBuf);

   If SizeOf(ExtBuf^) > MaxAvail then
      Abort(Msg);
   New(ExtBuf);
end {Initialize};

{ --------------------------------------------------------------------------- }

Procedure Read_Zip_Block;
Begin
   BlockRead(ZipFile, ZipBuf^, BUFSIZE, ZipCount);
   If ZipCount = 0 then
      EndFile := TRUE;
   ZipPtr := 1;
End {Read_Zip_Block};

{ --------------------------------------------------------------------------- }

Procedure Write_Ext_Block;
Begin
   If ExtPtr > 1 then begin
      BlockWrite(ExtFile, ExtBuf^, Pred(ExtPtr));
      If NOT IO_Test then
         Halt;
      ExtPtr := 1;
   end {if};
End {Write_Ext_Block};

{ --------------------------------------------------------------------------- }

Procedure Open_Zip;
Begin
   Assign(ZipFile, ZipName);
   FileMode := 64;  { Read Only / Deny None }
   {$I-} Reset(ZipFile, 1) {$I+};
   If NOT IO_Test then
      Halt;
   EndFile := FALSE;
   Read_Zip_Block;
End {Open_Zip};

{ --------------------------------------------------------------------------- }

Function Open_Ext : Boolean;
Begin
   Assign(ExtFile, OutPath + Hdr_FileName);
   FileMode := 66;  { Read & Write / Deny None }
   {$I-} Rewrite(ExtFile, 1) {$I+};
   If NOT IO_Test then
      Open_Ext := FALSE
   else begin
      ExtPtr := 1;
      Open_Ext := TRUE;
   end {if};
end {Open_Ext};

{ --------------------------------------------------------------------------- }

Function Get_Zip : Integer;
Begin
   If ZipPtr > ZipCount then
      Read_Zip_Block;

   If EndFile then
      Get_Zip := -1
   else begin
      Get_Zip := ZipBuf^[ZipPtr];
      Inc(ZipPtr);
   end {if};
end {Get_Zip};

{ --------------------------------------------------------------------------- }

Procedure Put_Ext(C : Byte);
Begin
   Crc32Val := UpdCrc(C, Crc32Val);
   ExtBuf^[ExtPtr] := C;
   Inc(ExtPtr);
   Inc(ExtCount);
   If ExtPtr > BUFSIZE then
      Write_Ext_Block;
end {Put_Ext};

{ --------------------------------------------------------------------------- }

Procedure Close_Zip;
Begin
   {$I-} Close(Zipfile) {$I+};
   If IO_Test then ;
end {Close_Zip};

{ --------------------------------------------------------------------------- }

Procedure Close_Ext;
Type
   TimeDateRec = Record
                    Time : Word;
                    Date : Word;
                 end {record};
Var
   TimeDate      : TimeDateRec;
   TimeDateStamp : LongInt Absolute TimeDate;
Begin
   Write_Ext_Block;
   TimeDate.Time := LocalHdr.Last_Mod_Time;
   TimeDate.Date := LocalHdr.Last_Mod_Date;
   SetFTime(ExtFile, TimeDateStamp);
   {$I-} Close(ExtFile) {$I+};
   If IO_Test then ;
   GotoXY(1, WhereY);
   Write(ExtCount);
   GotoXY(1, WhereY);
end {Close_Ext};

{ --------------------------------------------------------------------------- }

Procedure FSkip(Offset : LongInt);
Var
   Rec : LongInt;
Begin
   If (Offset + ZipPtr) <= ZipCount then
      Inc(ZipPtr, Offset)
   else begin
      Rec := FilePos(ZipFile) + (Offset - (ZipCount - ZipPtr) - 1);
      {$I-} Seek(ZipFile, Rec) {$I+};
      If NOT IO_Test then
         Halt;
      Read_Zip_Block;
   end {if};
end {FSkip};

{ --------------------------------------------------------------------------- }

Procedure FRead(Var Buf; RecLen : Word);
Var
   I : Word;
   B : Array[1..MaxInt] of Byte Absolute Buf;
Begin
   For I := 1 to RecLen do
      B[I] := Get_Zip;
end {FRead};

{ --------------------------------------------------------------------------- }

Function Read_Local_Hdr : Boolean;
Var
   Sign : LongInt;
Begin
   If EndFile then
      Read_Local_Hdr := FALSE
   else begin
      FRead(Sign, 4);
      If Sign = CENTRAL_FILE_HEADER_SIGNATURE then begin
         Read_Local_Hdr := FALSE;
         EndFile        := TRUE;
      end {then}
      else begin
         If Sign <> LOCAL_FILE_HEADER_SIGNATURE then
            Abort('Missing or invalid local file header in ' + ZipName);
         FRead(LocalHdr.Extract_Version_Reqd, SizeOf(LocalHdr) - 4);
         With LocalHdr do begin
            If FileName_Length > 255 then
               Abort('Filename of compressed file exceeds 255 characters!');
            FRead(Hdr_FileName[1], FileName_Length);
            Hdr_FileName[0] := Chr(FileName_Length);
            If Extra_Field_Length > 255 then
               Abort('Extra field of compressed file exceeds 255 characters!');
            FRead(Hdr_ExtraField[1], Extra_Field_Length);
            Hdr_ExtraField[0] := Chr(Extra_Field_Length);
         end {with};
         Read_Local_Hdr := TRUE;
      end {if};
   end {if};
end {Read_Local_Hdr};

{ --------------------------------------------------------------------------- }

Function Get_Compressed : Integer;
Var
   PctDone : Integer;
Begin
   If Bytes_To_Go = 0 then
      Get_Compressed := -1
   else begin
      Get_Compressed := Get_Zip;
      If (TenPercent = 0) OR (Bytes_To_Go MOD TenPercent = 0) then begin
         PctDone := 100 - Round(100.0 * (Bytes_To_Go / LocalHdr.Compressed_Size));
         GotoXY(WhereX - 4, WhereY);
         Write(PctDone:3, '%');
      end {if};
      Dec(Bytes_To_Go);
   end {if};
end {Get_Compressed};

{ --------------------------------------------------------------------------- }

Function LZW_Init : Boolean;
Var
   I : Word;
Label
   Exit;
Begin
   { Initialize LZW Table }
   If SizeOf(LZW_Table^) > MaxAvail then begin
      LZW_Init := FALSE;
      Goto Exit;
   end {if};
   New(LZW_Table);
   For I := 0 to TABLESIZE do begin
      With LZW_Table^[I] do begin
         Prefix     := -1;
         If I <= 255 then
            Suffix  := I
         else
            Suffix  := 0;
         ChildCount := 0;
      end {with};
   end {for};

   If SizeOf(FreeList^) > MaxAvail then begin
      LZW_Init := FALSE;
      Goto Exit;
   end {if};
   New(FreeList);
   For I := FIRSTENTRY to TABLESIZE do
      FreeList^[I] := I;
   NextFree := FIRSTENTRY;

   { Initialize the LZW Character Stack }
   If SizeOf(LZW_Stack^) > MaxAvail then begin
      LZW_Init := FALSE;
      Goto Exit;
   end {if};
   New(LZW_Stack);
   StackIdx := 0;

   LZW_Init := TRUE;

Exit:
end {LZW_Init};

{ --------------------------------------------------------------------------- }

Procedure LZW_Cleanup;
Begin
   If LZW_Table <> NIL then
      Dispose(LZW_Table);
   If FreeList <> NIL then
      Dispose(FreeList);
   If LZW_Stack <> NIL then
      Dispose(LZW_Stack);
end {LZW_Cleanup};

{ --------------------------------------------------------------------------- }

Procedure Clear_LZW_Table;
Var
   I : Word;
Begin
   StackIdx := 0;

   For I := FIRSTENTRY to TABLESIZE do begin           { Find all leaf nodes }
      If LZW_Table^[I].ChildCount = 0 then begin
         LZW_Stack^[StackIdx] := I;                  { and put each on stack }
         Inc(StackIdx);
      end {if};
   end {for};

   NextFree := Succ(TABLESIZE);

   While StackIdx > 0 do begin                        { clear all leaf nodes }
      Dec(StackIdx);
      I := LZW_Stack^[StackIdx];
      With LZW_Table^[I] do begin
         If Prefix <> -1 then
            Dec(LZW_Table^[Prefix].ChildCount);
         Prefix     := -1;
         Suffix     :=  0;
         ChildCount :=  0;
      end {with};
      Dec(NextFree);                         { add cleared nodes to freelist }
      FreeList^[NextFree] := I;
   end {while};
End {Clear_LZW_Table};

{ --------------------------------------------------------------------------- }

Procedure Add_To_LZW_Table(Prefix : Integer; Suffix : Byte);
Var
   I : Word;
Begin
   If NextFree <= TABLESIZE then begin
      I := FreeList^[NextFree];
      Inc(NextFree);
      LZW_Table^[I].Prefix     := Prefix;
      LZW_Table^[I].Suffix     := Suffix;
      Inc(LZW_Table^[Prefix].ChildCount);
   end {if};
End {Add_To_LZW_Table};

{ --------------------------------------------------------------------------- }

Function Get_Code(CodeSize : Byte) : Integer;
Const
   TmpInt     : Integer = 0;
Var
   BitsNeeded : Byte;
   HowMany    : Byte;
   HoldCode   : Integer;
Label
   Exit;
Begin
   If FirstCh then begin               { If first time through ...         }
      TmpInt := Get_Compressed;        { ... then prime the code buffer    }
      If TmpInt = -1 then begin        { If EOF on fill attempt ...        }
         Get_Code := -1;               { ... then return EOF indicator ... }
         Goto Exit;                    { ... and return to caller.         }
      end {if};
      SaveByte := TmpInt;
      BitsLeft := 8;                   { there's now 8 bits in our buffer  }
      FirstCh  := FALSE;
   end {if};

   BitsNeeded := CodeSize;
   HoldCode   := 0;

   While (BitsNeeded > 0) AND (TmpInt <> -1) do begin

      If BitsNeeded >= BitsLeft
         then HowMany := BitsLeft         { HowMany <-- Min(BitsLeft, BitsNeeded) }
         else HowMany := BitsNeeded;

      HoldCode := HoldCode OR ((SaveByte AND (1 SHL HowMany - 1)) SHL (CodeSize - BitsNeeded));
      SaveByte := SaveByte SHR HowMany;
      Dec(BitsNeeded, HowMany);
      Dec(BitsLeft, HowMany);

      If BitsLeft = 0 then begin          { If no bits left in buffer ...     }
         TmpInt := Get_Compressed;        { ... then attempt to get 8 more.   }
         If TmpInt = -1 then
            Goto Exit;
         SaveByte := TmpInt;
         BitsLeft := 8;
      end {if};

   end {while};

Exit:

   If BitsNeeded = 0 then             { If we got what we came for ... }
      Get_Code := HoldCode            { ... then return it             }
   else
      Get_Code := -1;                 { ... otherwise, return EOF      }

end {Get_Code};

{ --------------------------------------------------------------------------- }

Procedure UnShrink;
Var
   Ch       :  Char;
   CodeSize :  Byte;          { Current size (in bits) of codes coming in }
   CurrCode :  Integer;
   SaveCode :  Integer;
   PrevCode :  Integer;
   BaseChar :  Byte;
Label
   Exit;
Begin
   CodeSize := MINBITS;                   { Start with the smallest code size }

   PrevCode := Get_Code(CodeSize);        { Get first code from file          }
   If PrevCode = -1 then                  { If EOF already, then ...          }
      Goto Exit;                          { ... just exit without further ado }
   BaseChar := PrevCode;
   Put_Ext(BaseChar);                     { Unpack the first character        }

   CurrCode := Get_Code(CodeSize);        { Get next code to prime the while loop }

   While CurrCode <> -1 do begin          { Repeat for all compressed bytes   }

      If CurrCode = SPECIAL then begin    { If we've got a "special" code ... }

         CurrCode := Get_Code(CodeSize);
         Case CurrCode of
            1  :  Begin                   { ... and if followed by a 1 ...    }
                     Inc(CodeSize);       { ... then increase code size       }
                  end {1};
            2  :  Begin                   { ... and if followed by a 2 ...    }
                     Clear_LZW_Table;     { ... clear leaf nodes in the table }
                  end {2};
            else  begin                   { ... if neither 1 or 2, discard    }
                     Writeln;
                     Writeln('Encountered code 256 not followed by 1 or 2!');
                     Writeln;
                     Write('Press a key to continue ...');
                     Ch := ReadKey;
                     DelLine;
                     GotoXY(1, WhereY);
                  end {else};
         end {case};

      end {then}
      else begin                          { Not a "special" code              }

         SaveCode := CurrCode;            { Save this code someplace safe...  }

         If CurrCode > TABLESIZE then
            Abort('Invalid code encountered!');

         If (CurrCode >= FIRSTENTRY) AND (LZW_Table^[CurrCode].Prefix = -1) then begin
            If StackIdx > TABLESIZE then begin
               Write_Ext_Block;
               Abort('Stack Overflow!');
            end {if};
            LZW_Stack^[StackIdx] := BaseChar;
            Inc(StackIdx);
            CurrCode := PrevCode;         { Special case for KwKwK string     }
         end {if};

         While CurrCode >= FIRSTENTRY do begin
            If StackIdx > TABLESIZE then begin
               Write_Ext_Block;
               Abort('Stack Overflow!');
            end {if};
            LZW_Stack^[StackIdx] := LZW_Table^[CurrCode].Suffix;
            Inc(StackIdx);
            CurrCode := LZW_Table^[CurrCode].Prefix;
         end {while};

         BaseChar := LZW_Table^[CurrCode].Suffix;
         Put_Ext(BaseChar);               { Get last character ...            }

         While StackIdx > 0 do begin
            Dec(StackIdx);
            Put_Ext(LZW_Stack^[StackIdx]);
         end {while};                     { ... until there are none left     }

         Add_to_LZW_Table(PrevCode, BaseChar);  { Add new entry to table      }

         PrevCode := SaveCode;

      end {if};

      CurrCode := Get_Code(CodeSize);     { Get next code from input stream   }

   end {while};

Exit:
end {UnShrink};

{ --------------------------------------------------------------------------- }

Function Init_UnReduce : Boolean;
Label
   Exit;
Begin
   If SizeOf(Followers^) > MaxAvail then begin
      Init_UnReduce := FALSE;
      Goto Exit;
   end {if};
   New(Followers);

   If SizeOf(Stream^) > MaxAvail then begin
      Init_UnReduce := FALSE;
      Goto Exit;
   end {if};
   New(Stream);

   Init_UnReduce := TRUE;

Exit:
end {Init_UnReduce};

{ --------------------------------------------------------------------------- }

Procedure Cleanup_UnReduce;
Begin
   If Followers <> NIL then
      Dispose(Followers);
   If Stream <> NIL then
      Dispose(Stream);
end {Cleanup_UnReduce};

{ --------------------------------------------------------------------------- }

Function D(X, Y : Byte) : LongInt;
Begin
   X := X SHR (8 - Pred(LocalHdr.Compress_Method));
   D := X * 256 + Y + 1;
end {D};

{ --------------------------------------------------------------------------- }

Function F(X : Word) : Byte;
Begin
   If X = RedMask[Pred(LocalHdr.Compress_Method)] then
      F := 2
   else
      F := 3;
end {F};

{ --------------------------------------------------------------------------- }

Function L(X : Byte) : Byte;
Begin
   L := X AND RedMask[Pred(LocalHdr.Compress_Method)];
end {L};

{ --------------------------------------------------------------------------- }

Procedure StreamOut(C : Byte);
Begin
   Put_Ext(C);
   Stream^[StreamIdx] := C;
   StreamIdx := Succ(StreamIdx) AND 4095;
end {StreamOut};

{ --------------------------------------------------------------------------- }

Procedure ScrnchInit;
Begin
   State := 0;
   For StreamIdx := 0 to 4095 do
      Stream^[StreamIdx] := 0;
   StreamIdx := 0;
end {ScrnchInit};

{ --------------------------------------------------------------------------- }

Procedure UnScrnch(C : Byte);
Const
   DLE     =  $90;
Var
   S       :  LongInt;
   Count   :  Word;
   OneByte :  Byte;
Begin
   Case State of
      0  :  begin
               If C = DLE then
                  State := 1
               else
                  StreamOut(C);
            end {0};
      1  :  begin
               If C = 0 then begin
                  StreamOut(DLE);
                  State := 0;
               end {then}
               else begin
                  V     := C;
                  Len   := L(V);
                  State := F(Len);
               end {if};
            end {1};
      2  :  begin
               Inc(Len, C);
               State := 3;
            end {2};
      3  :  begin
               S := StreamIdx - D(V, C);
               If S < 0 then
                  S := S + 4096;
               Count := Len + 3;
               While Count > 0 do begin
                  OneByte := Stream^[S];
                  StreamOut(OneByte);
                  S := Succ(S) AND 4095;
                  Dec(Count);
               end {while};
               State := 0;
            end {3};
   end {case};
end {UnScrnch};

{ --------------------------------------------------------------------------- }

Function MinBits(Val : Byte) : Byte;
Begin
   Dec(Val);
   Case Val of
       0..1  : MinBits := 1;
       2..3  : MinBits := 2;
       4..7  : MinBits := 3;
       8..15 : MinBits := 4;
      16..31 : MinBits := 5;
      else     MinBits := 6;
   end {case};
end {MinBits};

{ --------------------------------------------------------------------------- }

Procedure UnReduce;
Var
   LastChar :  Byte;
   N        :  Byte;
   I, J     :  Word;
   Code     :  Integer;
Begin
   For I := 255 downto 0 do begin      { Load follower sets }
      N := Get_Code(6);                { Get size of 1st set }
      Followers^[I].SetSize := N;
      If N > 0 then
         For J := 0 to Pred(N) do
            Followers^[I].FSet[J] := Get_Code(8);
   end {for};

   ScrnchInit;

   LastChar := 0;
   Repeat

      If Followers^[LastChar].SetSize = 0 then begin
         Code := Get_Code(8);
         UnScrnch(Code);
         LastChar := Code;
      end {then}
      else begin
         Code := Get_Code(1);
         If Code <> 0 then begin
            Code := Get_Code(8);
            UnScrnch(Code);
            LastChar := Code;
         end {then}
         else begin
            I := MinBits(Followers^[LastChar].SetSize);
            Code := Get_Code(I);
            UnScrnch(Followers^[LastChar].FSet[Code]);
            LastChar := Followers^[LastChar].FSet[Code];
         end {if};
      end {if};
   Until (ExtCount = LocalHdr.Uncompressed_Size);
end {UnReduce};

{ --------------------------------------------------------------------------- }

Procedure UnZip;
Begin
   Crc32Val    := $FFFFFFFF;
   Bytes_To_Go := LocalHdr.Compressed_Size;
   FirstCh     := TRUE;

   ExtCount    := 0;

   TenPercent := LocalHdr.Compressed_Size DIV 10;

   Case LocalHdr.Compress_Method of
      0     :  Begin
                  While Bytes_to_go > 0 do
                     Put_Ext(Get_Compressed);
               end {0 = Stored};
      1     :  Begin
                  If LZW_Init then
                     UnShrink
                  else begin
                     Writeln('Not enough memory available to unshrink!');
                     Writeln('Skipping ', Hdr_FileName, ' ...');
                     FSkip(LocalHdr.Compressed_Size);
                     Crc32Val := NOT LocalHdr.Crc32;
                  end {if};
                  LZW_Cleanup;
               end {1 = shrunk};
      2..5  :  Begin
                  If Init_UnReduce then
                     UnReduce
                  else begin
                     Writeln('Not enough memory available to unreduce!');
                     Writeln('Skipping ', Hdr_FileName, ' ...');
                     FSkip(LocalHdr.Compressed_Size);
                     Crc32Val := NOT LocalHdr.Crc32;
                  end {if};
                  Cleanup_UnReduce;
               end {2..5};
      else     Begin
                  Writeln('Unknown compression method!');
                  Writeln('Skipping ', Hdr_FileName, ' ...');
                  FSkip(LocalHdr.Compressed_Size);
                  Crc32Val := NOT LocalHdr.Crc32;
               end {else};
   end {case};

   Crc32Val := NOT Crc32Val;
   If Crc32Val <> LocalHdr.Crc32 then begin
      Writeln;
      Writeln('WARNING: File ', OutPath + Hdr_FileName, ' fails CRC check!');
      Writeln('   Stored CRC = ', HexLInt(LocalHdr.Crc32),
              '   Calculated CRC = ', HexLInt(Crc32Val));
   end {if};
end {UnZip};

{ --------------------------------------------------------------------------- }

Procedure Extract_File;
Var
   YesNo : Char;
Label
   Exit;
Begin
   If FSearch(Hdr_FileName, OutPath) <> '' then begin
      Write('WARNING: ', OutPath + Hdr_FileName, ' already exists.  Overwrite (Y/N)? ');
      YesNo := ReadKey;
      Writeln(YesNo);
      If UpCase(YesNo) <> 'Y' then begin
         FSkip(LocalHdr.Compressed_Size);
         Goto Exit;
      end {if};
   end {if};

   If Open_Ext then begin
      Write('Extracting: ', OutPath + Hdr_FileName, ' ...    ');
      UnZip;
      GotoXY(WhereX - 4, WhereY);
      ClrEol;
      Writeln(' done');
      Close_Ext;
   end {then}
   else begin
      Writeln('Could not open output file ', OutPath + Hdr_FileName, '!  Skipping to next file ...');
      FSkip(LocalHdr.Compressed_Size);
   end {If};

Exit:
end {Extract_File};

{ --------------------------------------------------------------------------- }

Procedure Extract_Zip;
Var
   Match : Boolean;
   I     : Word;
Begin
   Open_Zip;
   While Read_Local_Hdr do begin
      Match := FALSE;
      I := 1;
      Repeat
         If SameFile(InFileSpecs[I], Hdr_FileName) then
            Match := TRUE;
         Inc(I);
      Until Match or (I > MaxSpecs);
      If Match then
         Extract_File
      else
         FSkip(LocalHdr.Compressed_Size);
   end {while};
   Close_Zip;
   GotoXY(1, WhereY);
   ClrEOL;
end;

{ --------------------------------------------------------------------------- }

Begin
   Assign(Output, '');
   Rewrite(Output);
   Writeln;
   Writeln(COPYRIGHT);
   Writeln(VERSION);
   Writeln;
   Load_Parms;   { get command line parameters }
   Initialize;   { one-time initialization }
   Extract_Zip;  { de-arc the file }
end.
