;.MODEL small
CGROUP group _TEXT
DGROUP group _DATA,STACK

COMMENT #

                 "AP-SLIDER"  (LZW) DYNAMIC DICTIONARY COMPRESSION
                              uses 12 bit pointers and
                                    4k dictionary
                                       w/update
                               (oldest leaf is deleted,
                                    one-at-a-time)
                                  Memory needed: 60k

                                 Written in A86 4.05
                                 Miles Pawski, 1992

                                THIS PROGRAM IS FREE

      To assemble & link under A86 4.05:  a86 +O +D +G14 kboom.asm
                                          qlink /DOSSEG

      Requirements:  8086+

      Please note:  .STARTUP generates the following code (if default
                    NEARSTACK is used - this places the stack in DGROUP so
                    ss:sp needs to be adjusted).  Only the first two lines
                    are required if using FARSTACK (stack in its own
                    segment).  FARSTACK can be specified by indicating
                    it in the .MODEL directive: .MODEL small,farstack

                                    mov dx,DGROUP
                                    mov ds,dx
                                    mov bx,ss
                                    sub bx,dx
                                    shl bx,1
                                    shl bx,1
                                    shl bx,1
                                    shl bx,1
                                    cli
                                    mov ss,dx
                                    add sp,bx
                                    sti

                    .EXIT generates terminate program code using int 21h,
                          function 04ch.

                    THIS PROGRAM is based on James Storer's 'squeeze' source
                    code found in his book DATA COMPRESSION, METHODS AND
                    THEORY, 1988 - Computer Science Press.  Modification
                    and methods were based on experimentation, general
                    tree theory of data structures and ideas put forth in
                    Mr. Storer's book.  Mr. Storer's original source code
                    is in PASCAL.  This program will hopefully provide a
                    better understanding of the LZW method and complex,
                    linked tree structures and their use.

 #


 ;---------------------------MACROS---------------------------------------

 child macro ;mainptr,char
   xor ax,ax                           ;returns pointer value in ax
   mov bx,#1
   mov cx,#2
   shl bx,1                            ;make mainptr word indexer
   mov ax,lcfield[bx]                  ;get child from lcfield
   mov bx,ax                           ;put in bx to index
   shl bx,1                            ;make word ptr
     M0:
     cmp ax,NILPTR
     je >M1
     cmp cx,ctrfield[bx]
     je >M1
       mov ax,rsibfield[bx]            ;cx=indexed char
       mov bx,ax                       ;if first child <> NILPTR and
       shl bx,1                        ;char <> indexed char then walk the
     jmp M0                            ;linked list of siblings of child
     M1:                               ;until we find a match
 #EM                                   ;no match returns NILPTR

 ctr macro ;mainptr
   xor ax,ax                           ;returns pointer value in ax
   mov bx,#1
   shl bx,1
   mov ax,ctrfield[bx]
 #EM

 parent macro ;mainptr
   xor ax,ax                           ;returns pointer value in ax
   mov bx,#1
   shl bx,1
   mov ax,parfield[bx]
 #EM

 writestring macro ;string
   mov ah,09h                          ;write string to screen
   lea dx,#1
   push ds
   push cs
   pop ds
   int 21h
   pop ds
 #EM

 writefile macro ;handle,var,len
   mov ah,40h                          ;write to file handle from variable
   mov bx,#1
   mov cx,#3
   lea dx,#2
   int 21h
 #EM

 readfile macro ;handle,var,len
   mov ah,3fh                          ;read from file handle into variable
   mov bx,#1
   mov cx,#3
   lea dx,#2
   int 21h
 #EM

 close macro ;handle
   mov ah,3eh                          ;file closure
   mov bx,#1
   int 21h
 #EM

;-------------------------------------------------------------------------
;   ADD_LEAF macro.  Adds a leaf to the tree.
;   Input: Parent node, node (code) to be added, character (mval)
;   ** See James Storer's  DATA COMPRESSION, METHODS AND THEORY
;      for more explanations, theory and original PASCAL code.
;-------------------------------------------------------------------------

 add_leaf macro
   cmp prevptr,NILPTR
   je >M1
   push si
   push di
   mov bx,trieptr                      ;new code
   mov cx,bx                           ;save a copy of new code in cx
   mov ax,mval                         ;match character
   mov di,prevptr                      ;parent node in di
   shl bx,1
   mov ctrfield[bx],ax                 ;ctrfield(mainptr)=char
   mov parfield[bx],di                 ;parfield(mainptr)=parptr
   mov lcfield[bx],NILPTR              ;lcfield(mainptr)=NILPTR
   mov lsibfield[bx],NILPTR
   shl di,1                            ;make parptr word indexer
   mov si,lcfield[di]                  ;si=lcfield(parptr)
   mov rsibfield[bx],si                ;rsibfield(mainptr)=lcfield(parptr)
   cmp si,NILPTR
   je >M0                              ;si=lcfield(parptr)
     shl si,1
     mov lsibfield[si],cx              ;lsibfield(lcfield(parptr))=mainptr
   M0:
   mov lcfield[di],cx                  ;lcfield(parptr)=ptr
   pop di
   pop si
   M1:
 #EM

;-------------------------------------------------------------------------
;  DELETE_LEAF macro.  Deletes a leaf from the tree.
;  Input: Node (code) to be deleted.  Passed in variable (trieptr).
;  If node has siblings, they will be connected to each other properly
;  after the node is deleted.
;  ** See James Storer's  DATA COMPRESSION, METHODS AND THEORY
;     for more explanations, theory and original PASCAL code.
;-------------------------------------------------------------------------

 delete_leaf macro
   push si
   push di
   mov bx,trieptr                      ;bx=code to be deleted
   shl bx,1
   mov ax,rsibfield[bx]                ;ax=rsibfield(mainptr)
   cmp lsibfield[bx],NILPTR
   je >M0
     mov di,lsibfield[bx]              ;connect left sibling and right sibling
     shl di,1                          ;through rsibfield of left sibling (huh?)
     mov rsibfield[di],ax              ;rsibfield[lsibfield(mainptr))=
     jmp >M1                           ;                 rsibfield(mainptr)
   M0:
     mov di,parfield[bx]               ;di=parfield(mainptr)
     shl di,1
     mov lcfield[di],ax                ;lcfield(parfield(mainptr))=
   M1:                                 ;                 rsibfield(mainptr)
   cmp ax,NILPTR
   je >M0
     mov di,lsibfield[bx]              ;di=lsibfield(mainptr)
     mov si,ax                         ;make left sibling of deleted leaf the
     shl si,1                          ;left sib. of deleted leaf's right sib.
     mov lsibfield[si],di              ;lsibfield[rsibfield(mainptr))=
   M0:                                 ;                 lsibfield(mainptr)
   pop di
   pop si
 #EM

;-------------------------------------------------------------------------

    ;Assembly-time constants

  MAXINCREMENT  EQU     3     ;AP - All Prefixes
  MAXMATCH      EQU   100     ;max string match
  STATICSIZE    EQU   256     ;constant dictionary guarantees first match
  DICTIONARY    EQU  4096
  BUFSIZE       EQU  8192
  STRSIZE       EQU    50
  MAXPTR        EQU  DICTIONARY-2    ;highest pointer
  NILPTR        EQU  DICTIONARY-1    ;end code - root code

    ;For experimentation.  If the DICTIONARY size is changed, this
    ;text sequence automatically changes CODESIZE and BITSHIFT to fit
    ;the DICTIONARY.  CODESIZE and BITSHIFT are used in the code to
    ;read and write pointers from and to the bit stream.
    ;Use only the below DICTIONARY values.

  #IF DICTIONARY EQ 512
     CODESIZE EQU 9
     BITSHIFT EQU 7
  #ELSEIF DICTIONARY EQ 1024
     CODESIZE EQU 10
     BITSHIFT EQU 6
  #ELSEIF DICTIONARY EQ 2048
     CODESIZE EQU 11
     BITSHIFT EQU 5
  #ELSEIF DICTIONARY EQ 4096
     CODESIZE EQU 12
     BITSHIFT EQU 4
  #ENDIF

_DATA SEGMENT WORD PUBLIC 'DATA' ;.DATA?

EVEN 2
 inputbyte         dw     ?
 dictsize          dw     ?
 fstadd            db     ?
 bigplace          dw     ?
 curptr            dw     ?
 curlen            dw     ?
 prevptr           dw     ?
 prevlen           dw     ?
 trieptr           dw     ?
 match             db     MAXMATCH dup (?),?
 mstart            dw     ?
 mval              dw     ?
 mindex            dw     ?

 ;-------------------------------------------
 fielder    label  WORD

 ctrfield          dw     DICTIONARY dup (?)
 lcfield           dw     DICTIONARY dup (?)
 parfield          dw     DICTIONARY dup (?)
 rsibfield         dw     DICTIONARY dup (?)
 lsibfield         dw     DICTIONARY dup (?)
 ;-------------------------------------------

 filename          db     STRSIZE dup (?)
 filename2         db     STRSIZE dup (?)
 parabuf           db     STRSIZE dup (?)
 ascii_string      db     10 dup (?)
 work_code         db     4 dup (?)
 init_size         dd     ?
 comp_size         dd     ?
 xdate             dw     ?
 xtime             dw     ?

 outbuffer         db     BUFSIZE dup (?)   ;file i/o buffers
 inbuffer          db     BUFSIZE dup (?)

;_DATA ENDS
STACK SEGMENT WORD STACK 'STACK' ;.STACK

  dw 1024 dup (?)
  top_stack label WORD

;STACK ENDS
_TEXT SEGMENT BYTE PUBLIC 'CODE' ;.CODE

jmp MAIN

  incount       dw  0
  bytes_read    dw  0
  outcount      dw  0
  master_size   dd  0
  sixteen_count db  16
  paralen       db  0
  nologo_flag   db  0
  comp_flag     db  0
  decomp_flag   db  0
  end_flag      db  0
  all_over      db  0
  file_handle   dw  0
  file_handle2  dw  0
  filelen       dw  0
  filelen2      dw  0

  kboom_code    db  'MP'
  comp          db  'Compressing: $'
  expand        db  'Expanding: $'
  arrow         db  ' --> $'
  isize         db  'Orig. size: $'
  csize         db  'Comp. size: $'
  ratio         db  'Ratio: $'
  fsize         db  'Exp. size: $'
  crlf          db  13,10,'$'
  err           db  'I/O Error$'
  sorry         db  'Sorry...$'
  sorry2        db  ' was not compressed by Kboom11$'

  logo          db  13,10,'  KBoom11  LZW File Encoding/Decoding Utility featuring',13,10
                db        '  4k Dynamic Dictionary w/discrete update',13,10
                db        '  Miles Pawski  CIS: 70473,527       1992',13,10,13,10,'$'

  format        db  'FORMAT: kboom [input file] [output file] [/c][/d][/n]',13,10
                db  '        /c - Compression /d - Decompression /n - No logo$'

MAIN: ;.STARTUP

 ;mov ax,seg _DATA
 mov ax,DGROUP
 mov ds,ax
 mov bx,ss
 sub bx,ax
 mov cx,4
 shl bx,cl
 cli
 mov ss,ax
 add sp,bx
 ;mov sp,offset (DGROUP:top_stack)
 sti

;-----------------------------MAIN PROGRAM--------------------------------

 mov bp,sp
 push offset parabuf
 push offset filename2
 push offset filename
 call cmdline
 ;add sp,6
 mov sp,bp
 xor ax,ax
 mov al,cs:[paralen]
 cmp ax,0                              ;if no parameters, quit with
 je >L8                                ;FORMAT display

    push ax
    push offset parabuf
    call caps                          ;capitalize parameters
    pop si
    pop cx

    dec cx
    L0:                                ;read parameters
      mov ax,word ptr [si]
      cmp ax,'N/'
      jne >L1
        mov cs:[nologo_flag],1
        jmp >L3
      L1:
      cmp ax,'C/'
      jne >L2
        mov cs:[comp_flag],1
        jmp >L3
      L2:
      cmp ax,'D/'
      jne >L3
        mov cs:[decomp_flag],1
      L3:
      inc si
    loop L0

    call initialize                    ;set variables

    cmp cs:[comp_flag],1
    jne >L0
      call compression
      jmp >L9
    L0:
    cmp cs:[decomp_flag],1
    jne >L8
      call decompression
      jmp >L9

 L8:
   writestring format

 L9:
   cmp cs:[file_handle],0
   je >L0
     close cs:[file_handle]
     close cs:[file_handle2]
   L0:
   ;.EXIT
   mov ah,4ch
   int 21h

;-------------------------------------------------------------------------


;ͻ
;                                                                      
;                                                                      
;                           PROCEDURE SECTION                          
;                                                                      
;                                                                      
;ͼ

;------------------------------------------------------------------------
;                                                                       |
;                    MAIN COMPRESSION SEQUENCE                          |
;                                                                       |
;------------------------------------------------------------------------

 compression proc near

  call open_source_file                ;open source file
  IF C jmp fincp

  mov ah,4eh                           ;"first match" interrupt
  mov cx,20h                           ;file attribute
  mov dx,offset filename               ;point to filespec
  int 21h

  push es
  mov ah,2fh                           ;get dta
  int 21h
  mov ax,word ptr es:[bx+1ah]          ;get file size from dta
  mov dx,word ptr es:[bx+1ah][2]
  mov word ptr [init_size][0],ax       ;record in variable
  mov word ptr [init_size][2],dx
  mov ax,word ptr es:[bx+16h]          ;get file time
  mov xtime,ax
  mov ax,word ptr es:[bx+18h]          ;get file date
  mov xdate,ax
  pop es

  call create_second_file              ;create target file
  IF C jmp fincp

  cmp cs:[nologo_flag],1
  je >L0
    writestring logo
  L0:

  writestring comp                     ;"Compressing: "
  writefile 1,filename,cs:[filelen]    ; [filename]
  writestring arrow
  writefile 1,filename2,cs:[filelen2]

  mov si,offset kboom_code
  mov di,offset work_code
  mov ax,ptr word cs:[si]
  mov ptr word [di],ax
  add si,2
  add di,2
  mov ax,ptr word cs:[si]
  mov ptr word [di],ax
  writefile cs:[file_handle2],work_code,4 ;write kboom code to file in
                                       ;first 4 bytes so we won't try to
                                       ;expand any files not generated by
                                       ;this program

  writefile cs:[file_handle2],init_size,4 ;write next 4 bytes to file
                                       ;which represent what file size
                                       ;should be when expansion is
                                       ;finished

  writefile cs:[file_handle2],xdate,2  ;write date to file
  writefile cs:[file_handle2],xtime,2  ;write time to file

  mov di,offset outbuffer              ;set output buffer

  call read_file                       ;read file into buffer
  lodsb                                ;get first byte from buffer
  inc cs:[incount]
  xor ah,ah                            ;convert to word
  mov inputbyte,ax                     ;move byte into variable

  mov mstart,1


              ;--------- MAIN COMPRESSION LOOP ----------------------

  L0:                                  ;loop indefinitely

    mov ax,inputbyte                   ;same as: child NILPTR, inputbyte
    mov trieptr,ax                     ;First ptr is always character and
                                       ;gets us into the root level

    mov ax,curptr
    mov prevptr,ax                     ;prevptr=curptr
    mov ax,curlen
    mov prevlen,ax                     ;prevlen=curlen
    mov curlen,0


          L1:
          cmp trieptr,NILPTR
          je leaving                   ;first trieptr is never NILPTR

            mov ax,trieptr
            mov curptr,ax              ;curptr=trieptr
            mov ax,inputbyte
            inc curlen
            mov bx,curlen
            mov match[bx],al           ;Record byte in array starting
                                       ;at position 1.  Curlen indexes in.
                                       ;This is used by update process.
                                       ;First byte in loop is always a match

                mov ax,cs:[incount]
                cmp ax,cs:[bytes_read] ;See if we're at end of input buffer.
                IF AE jmp read_leave   ;If not, keep reading

            goon:
                lodsb                  ;get new byte
                inc cs:[incount]
                xor ah,ah
                mov inputbyte,ax       ;place in variable
                child curptr,inputbyte ;see if new byte is child of curptr
                mov trieptr,ax         ;result in ax -> trieptr

          jmp L1                       ;keep walking down the tree if
                                       ;we have a match (trieptr <> NILPTR)

          leaving:
            xor ax,ax
            mov al,cs:[sixteen_count]  ;get number of bit positions left

            mov bx,curptr              ;Move curptr in bx.
            shl bx,BITSHIFT            ;CODESIZE bit pointer will now start
                                       ;at left end of word.  We want to
                                       ;output curptr into bit stream
            xor cx,cx
            cploop:
              cmp cx,CODESIZE          ;Check code bit counter.
              je outcp                 ;Done?  Then leave.  Shift both in
              shl bx,1                 ;unison.  dx holds bits.
              rcl dx,1                 ;Carry added to dl if set by bh.
              inc cx                   ;We are copying bx to dx bit by
              dec ax                   ;bit. ax=dx bit counter
            jnz cploop

          ;---------------- JUMP-OUTS & FALL-THROUGHS ----------------------
          ;                   conditional branches                         ;
          ;             kept out of main loop for speed                    ;
          ;                                                                ;
              mov ax,dx                                                    ;
              stosw                       ;store word in buffer            ;
              xor dx,dx                   ;clear out dx                    ;
              mov ax,16                   ;reset output bit counter        ;
                add cs:[outcount],2                                        ;
                cmp cs:[outcount],BUFSIZE ;See if we're at end of output   ;
                jb >L1                    ;buffer.  If not, go on.         ;
                  call write_file         ;If at end, flush the buffer     ;
                                                                           ;
              L1:                                                          ;
              cmp cx,CODESIZE-8           ;If we have enough bits left     ;
              ja cploop                   ;and dx is clear, then we can    ;
              mov dl,bh                   ;move a whole byte (bh) at once. ;
              mov bh,bl                   ;move bl over to left of word    ;
              mov ax,8                    ;adjust output bit counter       ;
              add cx,8                    ;adjust bit counter for CODESIZE ;
              jmp cploop                                                   ;
                                                                           ;
              read_leave:                                                  ;
              cmp cs:[end_flag],1                                          ;
              jne >L1                     ;Check for eof.  end_flag set if ;
                mov cs:[all_over],1       ;eof.  Set another flag so we    ;
                jmp leaving               ;can process the last byte       ;
                                                                           ;
              L1:                                                          ;
              call read_file              ;read more from file into buffer ;
              jmp goon                                                     ;
          ;                                                                ;
          ;-----------------------------------------------------------------


          outcp:
            mov cs:[sixteen_count],al  ;save bit count
            cmp cs:[all_over],1        ;Last byte?  Then leave
            je endcp

            call update                ;update the tree

  jmp L0

              ;-------------- END MAIN COMPRESSION LOOP -------------

  endcp:

  xor cx,cx
  mov cl,cs:[sixteen_count]
  cmp cx,16                            ;See if we have leftover in dx.
  jae >L0                              ;If so, shift them all to the
    shl dx,cl                          ;left so we don't mess up the
    mov ax,dx                          ;bit stream
    stosw                              ;record final bytes
    add cs:[outcount],2
  L0:

  call write_file                      ;flush the buffer


              ;--------- INFO/DECORATION ----------------------------

  writestring crlf                     ; carriage return-line feed
  writestring isize                    ;"Orig. size: "
  mov ax,word ptr init_size[0]         ;Convert to decimal string.
  mov dx,word ptr init_size[2]         ;Value is passed to process in
  call convert                         ;dx:ax pair


  writestring crlf
  writestring csize                    ;"Comp. size: "
  mov ax,word ptr cs:[master_size][0]
  mov dx,word ptr cs:[master_size][2]
  call convert

  writestring crlf
  writestring ratio                    ;"Ratio: "

                        ;Ratio is not % saved but, rather
                        ;the new file length divided by the old.
                        ;However, to achieve this, I multiply the
                        ;new file length by 100 first, then divide
                        ;it by the old length

  mov si,100
  xor dx,dx
  mov ax,word ptr cs:[master_size][0] ;file length: get low word
  mul si                          ;multiply by 100
  mov cx,dx                       ;save dx in cx
  mov bx,ax                       ;save low word in bx
  xor dx,dx
  mov ax,word ptr cs:[master_size][2] ;get high word              dx   ax
  mul si                          ;multiply by 100                     si
  add dx,cx                       ;add saved high word          ---------
  add dx,ax                       ;add high to high  (si*ax)      cx   bx
  mov ax,bx                       ;get back low      (si*dx)  dx  ax
                                  ;                         -------------
                                  ;Oh well ->                 dx+ax+cx,bx

  mov bx,word ptr init_size[0]
  mov cx,word ptr init_size[2]

                        ;dx:ax now has the new file size multiplied by 100.
                        ;cx:bx has original file size


              ;---------DIVISION ROUTINE-----------------------------

  L0:
  or      cx,cx                   ;Divide both numbers by 2
  jz      >L1                     ;until high word of
  shr     dx,1                    ;divisor (cx) is zero.
  rcr     ax,1                    ;If cf is set by dx shift, the bit
                                  ;will be added to the left of ax during
                                  ;ax rotate
  shr     cx,1
  rcr     bx,1
  jmp     L0

  L1:
  push    ax                      ;save low word
  mov     ax,dx
  xor     dx,dx
  div     bx                      ;divide high word
  mov     cx,ax                   ;save high quotient
  pop     ax
  add     ax,dx                   ;add remainder to low word
  div     bx                      ;divide low word
  xchg    cx,dx                   ;save remainder in cx
                                  ;386 code is faster and easier and I would
                                  ;much rather use it than going through all
                                  ;this BS.  32-bit registers FOREVER!

  push dx                         ;save high word
  push ax                         ;save low word
  xor dx,dx                       ;zero dx
  mov ax,cx                       ;get back last remainder
  mov si,10
  mul si                          ;multiply by 10

  div bx                          ;divide by divisor again
  mov cx,ax                       ;save last quotient in cx
                                  ;this will be first digit after decimal

  mov ax,dx                       ;put latest remainder into ax
  xor dx,dx                       ;zero dx
  mul si                          ;multiply by 10
  div bx                          ;divide by divisor
                                  ;this will be second digit after decimal
  mov bx,ax                       ;save quotient in bx

              ;------- END DIVISION ROUTINES ------------------------

  pop ax                          ;get back low word
  pop dx                          ;get back high word

  push bx                         ;save second decimal digit
  push cx                         ;save first decimal digit
  call convert                    ;write main ratio
  pop cx                          ;get back first decimal digit
  pop bx                          ;get back second decimal digit

  mov ah,02h
  mov dl,'.'                      ;write decimal point "."
  int 21h

  mov dl,cl                       ;put first digit into dl
  add dl,30h                      ;convert to ascii
  int 21h                         ;write first digit after decimal to screen

  mov dl,bl                       ;put second digit into dl
  add dl,30h                      ;convert to ascii
  int 21h                         ;write second digit after decimal to screen

  mov dl,'%'                      ;"%"
  int 21h

              ;---------- END INFO DECORATION -----------------------

  fincp:
    ret

 compression endp

;------------------------------------------------------------------------
;                                                                       |
;                    MAIN DECOMPRESSION SEQUENCE                        |
;                                                                       |
;------------------------------------------------------------------------

 decompression proc near

  call open_source_file                ;open source file
  IF C jmp findp

  mov ah,4eh                           ;"first match" interrupt
  mov cx,20h                           ;file attribute
  mov dx,offset filename               ;point to filespec
  int 21h

  push es
  mov ah,2fh
  int 21h                              ;get dta
  mov ax,word ptr es:[bx+1ah]          ;get file size from dta
  mov dx,word ptr es:[bx+1ah][2]
  sub ax,12                            ;Subtract 12 bytes because file size
  sbb dx,0                             ;and code are contained in these bytes.
  mov word ptr [comp_size][0],ax       ;Record compressed file size (from dta).
  mov word ptr [comp_size][2],dx       ;Date and time also take 4 bytes
  pop es

  readfile cs:[file_handle],work_code,4 ;first four bytes are compression code

  mov si,offset work_code              ;check to see if code is kboom
  mov di,offset kboom_code
  mov ax,word ptr [si]
  mov bx,word ptr cs:[di]
  cmp ax,bx                            ;check first pair
  jne >L0
    add si,2
    add di,2
    mov ax,word ptr [si]
    mov bx,word ptr cs:[di]
    cmp ax,bx                          ;if ok, check second pair
    je >L1
  L0:
    writestring sorry
    writefile 1,filename,cs:[filelen]
    writestring sorry2
    jmp findp
  L1:

  call create_second_file              ;open target file
  IF C jmp findp

  cmp cs:[nologo_flag],1
  je >L0
    writestring logo
  L0:

  writestring expand                   ;"Expanding [] --> []"
  writefile 1,filename,cs:[filelen]
  writestring arrow
  writefile 1,filename2,cs:[filelen2]

  writestring crlf
  readfile cs:[file_handle],init_size,4 ;second four bytes are file size
  readfile cs:[file_handle],xdate,2    ;read file date into variable
  readfile cs:[file_handle],xtime,2    ;read file time into variable

  mov di,offset outbuffer

  call read_file                       ;read file into buffer
  lodsw
  mov dx,ax                            ;load dx with first word
  add cs:[incount],2


              ;--------------------MAIN DECOMPRESSION ROUTINE--------

  L0:                                  ;loop indefinitely

    xor ax,ax
    mov al,cs:[sixteen_count]          ;get saved bit count
    xor bx,bx                          ;We need to read CODESIZE bits from input stream.
    xor cx,cx                          ;We read a bit at a time from dx into bx.  The
    dploop:                            ;leftover is left in dx for next time.

      cmp cx,CODESIZE                  ;See if CODESIZE bits have been read.
      je outdp                         ;If not, read more bits.
                                       ;If so, we have full CODESIZE bit pointer. Leave.
      shl dx,1
      rcl bx,1                         ;bx will have pointer.  Shift in unison with
      inc cx                           ;dx.  ax keeps bit count of dx.  If dx is empty,
      dec ax                           ;fall through and fill dx again.
    jnz dploop

   ;-------------------JUMP-OUTS & FALL-THROUGHS ----------------------------
   ;                                                                        ;
       mov ax,cs:[incount]        ;Check byte count.  See if all bytes have ;
       cmp ax,cs:[bytes_read]     ;been processed from buffer.              ;
       jb check_end               ;If not, proceed                          ;
         cmp cs:[end_flag],1                                                ;
         jne >L1                                                            ;
           mov cs:[all_over],1                                              ;
           jmp outdp                                                        ;
         L1:                                                                ;
         call read_file           ;read more of file into buffer            ;
                                                                            ;
       check_end:                 ;jump-out from above loop                 ;
       lodsw                                                                ;
       mov dx,ax                  ;put new word into dx                     ;
       add cs:[incount],2         ;count it                                 ;
       mov ax,16                  ;reset bit count                          ;
       jmp dploop                                                           ;
    ;                                                                       ;
    ;------------------------------------------------------------------------


    outdp:
    mov cs:[sixteen_count],al          ;save bit count
    mov trieptr,bx                     ;put completed pointer into trieptr
                                       ;GOT POINTER, NOW START
    mov ax,curptr
    mov prevptr,ax                     ;prevptr=curptr
    mov ax,trieptr
    mov curptr,ax                      ;curptr=trieptr
    mov ax,curlen
    mov prevlen,ax                     ;prevlen=curlen
    mov curlen,0


          L1:
          cmp trieptr,NILPTR
          je >L2
             ctr trieptr               ;get character for trieptr (ret in ax)
             mov bx,MAXMATCH
             sub bx,curlen             ;Index to end of match array.  Place
             mov match[bx],al          ;character there.  Match array is
             inc curlen                ;written back to front.
             mov cx,bx
             parent trieptr            ;Get parent of trieptr (ret in ax).
             mov trieptr,ax            ;Move into trieptr.  Walk up the tree
          jmp L1                       ;reading the nodes until we reach the
          L2:                          ;root (NILPTR)


          mov bx,cx
          mov mstart,cx                ;Move index into mstart.  This is used
                                       ;by update routine.

          L1:
          cmp bx,MAXMATCH
          ja >L2                       ;MAXMATCH length is 100 letters
             mov al,match[bx]          ;write the string in the match array
             stosb                     ;to the output buffer
             inc cs:[outcount]
             cmp cs:[outcount],BUFSIZE ;Buffer holds BUFSIZE.  When full,
             jb check_end2             ;flush buffer.  Do this outside of
                call write_file        ;loop.  Hate to put jump-arounds
                                       ;inside loop.  Fall-through is faster
             check_end2:               ;jump-out from above loop
             inc bx
          jmp L1
          L2:
          cmp cs:[all_over],1
          je enddp

          call update

  jmp L0


  enddp:

  call adjust
  call write_file

  writestring csize                    ;"Comp. size: "
  mov ax,word ptr comp_size[0]
  mov dx,word ptr comp_size[2]
  call convert

  writestring crlf
  writestring fsize                    ;"Exp. size: "
  mov ax,word ptr cs:[master_size][0]
  mov dx,word ptr cs:[master_size][2]
  call convert


  mov ah,57h
  mov al,01h
  mov bx,cs:[file_handle2]             ;set date and time
  mov cx,xtime
  mov dx,xdate
  int 21h

  findp:
    ret

 decompression endp


;--------------------------------------------------------------------
; INITIALIZE procedure.  Initializes variables and the first 256 nodes
; of the tree.  The first 256 nodes represent the first level of the
; tree and they will never be deleted in this particular data structure.
;--------------------------------------------------------------------

   initialize proc near
     mov di,ds
     mov es,di

     mov ax,NILPTR
     mov di,offset fielder             ;fill fields with NILPTR
     mov cx,DICTIONARY*5
     rep stosw

     mov dictsize,0
     xor bx,bx
     L0:
     cmp dictsize,STATICSIZE
     jae >L1                           ;Fill first 256 bytes of ctr
       mov ax,dictsize                 ;(character) field with usual
       mov ctrfield[bx],ax             ;dos characters 0-255.
       add bx,2                        ;These will comprise the non-
       inc dictsize                    ;erasable portion of the dictionary
     jmp short L0
     L1:

     mov bigplace,STATICSIZE
     mov curptr,NILPTR
     mov curlen,0

     ret
   initialize endp


;--------------------------------------------------------------------
; UPDATE procedure.  Updates the tree by adding and deleting appropriate
; leaves.  Also determines which codewords will be re-used and which
; leaves will be deleted from the tree.
; Input: From main program - trieptr, prevptr, match, etc.
;--------------------------------------------------------------------

   update proc near

     cmp prevptr,NILPTR
     IF E jmp finup

       mov fstadd,0
       mov mindex,0


         L0:                           ;loop indefinitely

             mov ax,curlen             ;Don't loop more than current
             cmp mindex,ax             ;match length.  Leaf nodes are
             IF AE jmp finup           ;added to previous pointer

             cmp mindex,MAXINCREMENT   ;MAXINCREMENT is number of pre-
             IF AE jmp finup           ;fixes to add.  Read on...

             mov ax,prevlen            ;Maximum match length is 100
             add ax,mindex             ;in this code.  This routine
             cmp ax,MAXMATCH           ;adds prefixes one-at-a-time
             IF AE jmp finup           ;per loop.  Variable mindex is
                                       ;the loop counter

           mov bx,mstart               ;mstart from main program
           add bx,mindex               ;index into match array
           xor ah,ah                   ;make into word
           mov al,match[bx]            ;get letter from array
           mov mval,ax                 ;record in mval

           cmp fstadd,1                ;Skip the child search after first
           je >L1                      ;leaf added since the rest will be
                                       ;new and without children

           child prevptr,mval          ;see if mval is child of prevptr
           mov trieptr,ax              ;record result in trieptr

                             ;If no match was found, we will
                             ;walk down the tree from prevptr and
                             ;add leaves consisting of the
                             ;matched prefixes.  The number
                             ;added is regulated by MAXINCREMENT

             cmp trieptr,NILPTR
             IF NE jmp long >L3        ;if not child, then update tree
             mov fstadd,1
             L1:

               cmp dictsize,MAXPTR
               ja >L1
                 mov ax,dictsize       ;get free unused pointer
                 mov trieptr,ax        ;move pointer into trieptr
                 inc dictsize          ;increment to next available pointer
                 jmp >L2
               L1:                     ;No more unused pointers?
             mov cx,bigplace           ;Get saved place.  Will loop around
             bigtop:                   ;from 256 to NILPTR progressively and
               cmp cx,prevptr          ;re-use pointers.  Don't use prevptr.
               je around               ;We'll get the oldest leaf without
               cmp cx,MAXPTR           ;children, delete the leaf, update the
               ja bigtop2              ;tree with recycled pointer.  Over max?
             bigtop1:                  ;Then reset to 256 and start over
               mov bx,cx               ;put into bx
               shl bx,1                ;word index
               cmp lcfield[bx],NILPTR  ;See if pointer has children.
                                       ;NILPTR means no children for this
               je bigtopout            ;node and we can erase & re-use it,
             around:                   ;else keep trying
               inc cx                  ;go to next pointer
               jmp bigtop              ;repeat

             bigtop2:                  ;jump-out from above loop
               mov cx,STATICSIZE       ;reset to 256
               jmp bigtop1

             bigtopout:
               mov trieptr,cx          ;put leaf node in trieptr
               inc cx                  ;advance for progressive usage
               mov bigplace,cx         ;Record in bigplace.  We'll start
                                       ;here next time we look for a pointer
                                       ;to re-use
             delete_leaf               ;Delete leaf node and, if necessary,
                                       ;reconnect siblings
               L2:

             add_leaf                  ;add leaf to tree

             L3:

           mov ax,trieptr
           mov prevptr,ax
           inc mindex

         jmp L0

     finup:
       ret

   update endp


;--------------------------------------------------------------------
; ADJUSTMENT procedure.  The compression routine may add on an extra
; byte or two to the end.  This is usually the letter coded as 0 since
; the encoder has to flush the final byte to the file, whether it is
; full or not.  These extra bytes are removed by this routine.  The
; decoder never generates fewer bytes than what the final size
; of the file should be.  The encoder writes the original file
; size to the file at bytes 5-8.  The decoder keeps track of the
; number of bytes generated and compares it with this figure.
; This routine then subtacts the extra byte(s) off the end.  Don't
; worry, they're never needed and are really extra bytes. This
; is one way to do things and it works without fail.
;--------------------------------------------------------------------

   adjust proc near
     mov dx,word ptr cs:[master_size][2] ;load high word of byte count
     mov ax,word ptr cs:[master_size][0] ;load low word of byte count
     add ax,cs:[outcount]              ;add the last count
     adc dx,0                          ;add the carry flag to high word

     cmp dx,word ptr [init_size][2]    ;See if high word of byte count is
     je >L0                            ;equal to high word of header info.
     cmp dx,word ptr [init_size][2]    ;See if high word of byte count is
     jb >L1                            ;lower than high word of header info.
     sub ax,word ptr [init_size][0]    ;If so, leave without adjusting,
     sbb dx,word ptr [init_size][2]    ;otherwise, if byte count is greater
     sub cs:[outcount],ax              ;than header info, subtact the two
     jmp >L1                           ;and take the difference off the end
                                       ;so that the file size will be correct
     L0:
       cmp ax,word ptr [init_size][0]  ;Compare low words.
       jbe >L1                         ;If expanded file is smaller, leave
       sub ax,word ptr [init_size][0]  ;without adjusting, otherwise, subtract
       sub cs:[outcount],ax            ;the two and take the difference off
                                       ;the end again
     L1:
     ret
   adjust endp


;--------------------------------------------------------------------
; OPEN_SOURCE_FILE procedure.
; Input: None
; Output: Carry flag set if error
;--------------------------------------------------------------------

   open_source_file proc near
     mov si,offset filename
     mov bx,cs:[filelen]
     mov byte ptr [si][bx],0
     mov ah,3dh                        ;open input file
     mov al,00h                        ;read-only mode
     mov dx,si
     int 21h
     jnc >L0                           ;check to see if carry flag is set
       writestring err
       stc
       jmp >L1
     L0:
     mov cs:[file_handle],ax
     L1:
     ret
   open_source_file endp


;--------------------------------------------------------------------
; CREATE_OUTPUT_FILE procedure.
; Input: None
; Output: Carry flag set if error
;--------------------------------------------------------------------

   create_second_file proc near
     mov si,offset filename2
     mov bx,cs:[filelen2]
     mov byte ptr [si][bx],0
     mov ah,3ch                        ;create/truncate output file
     mov cx,00h                        ;normal attribute
     mov dx,si
     int 21h
     jnc >L0                           ;check to see if carry flag is set
       writestring err
       stc
       jmp >L1
     L0:
     mov cs:[file_handle2],ax
     L1:
     ret
   create_second_file endp


;--------------------------------------------------------------------
; READ_FILE procedure.  Reads the file into a buffer for processing
; by the program.
; Input: None
; Output: end_flag
;--------------------------------------------------------------------

   read_file proc near
     push ax
     push bx
     push cx
     push dx
     mov si,offset inbuffer
     mov cs:[incount],0
     mov ah,3fh
     mov bx,cs:[file_handle]
     mov cx,BUFSIZE
     mov dx,si
     int 21h
     mov cs:[bytes_read],ax            ;Number of bytes read will be in ax.
     cmp ax,BUFSIZE                    ;Less than requested bytes indicates
     jae >L0                           ;eof
       mov cs:[end_flag],1
     L0:
     pop dx
     pop cx
     pop bx
     pop ax
     ret
   read_file endp


;--------------------------------------------------------------------
; WRITE_FILE procedure.  Writes the output buffer to disk.
; Input: None
; Output: master_size
;--------------------------------------------------------------------

   write_file proc near
     push ax
     push bx
     push cx
     push dx
     mov cx,cs:[outcount]
     mov di,offset outbuffer
     mov cs:[outcount],0
     add word ptr cs:[master_size][0],cx ;Keeps track of bytes written
     adc word ptr cs:[master_size][2],0 ;to file in variable master_size.
     mov ah,40h                        ;This variable will be referenced
     mov bx,cs:[file_handle2]          ;later by adjustment routine in
     mov dx,di                         ;the decoder.  Encoder doesn't
     int 21h                           ;use this variable so no harm done
     pop dx                            ;there
     pop cx
     pop bx
     pop ax
     ret
   write_file endp


;--------------------------------------------------------------------
; CMDLINE procedure.  Parses the command line for files.
;--------------------------------------------------------------------

   cmdline proc near

     push bp
     mov bp,sp

     ;mov ah,51h              ;get psp
     ;int 21h
     ;mov es,bx               ;move psp seg into es
     mov bx,80h
     xor cx,cx
     xor dx,dx               ;selector flag=0
     cmp es:[bx],cl          ;Check for command line arguments.
     je >L5                  ;Value at 80h of the psp is number of bytes
                             ;in command line argument

       add bx,2                   ;find start of command line
       mov si,[bp+4]              ;buffer for input file
       L0:
         mov al,es:[bx]           ;get letter from dta (command line)
         cmp al,13
         je >L4
         cmp al,0
         je >L4
         cmp cl,STRSIZE
         je >L5
         cmp al,32
         ja >L2
         cmp cl,0
         je >L3
         cmp dx,0
         ja >L1
           mov si,[bp+6]               ;buffer for output file
           mov cs:[filelen],cx
           xor cx,cx
           inc dx
           jmp >L3
         L1:
         cmp dx,1
         ja >L2
           mov si,[bp+8]               ;buffer for parameter line
           mov cs:[filelen2],cx
           xor cx,cx
           inc dx
           jmp >L3
         L2:
         mov [si],al
         inc si
         inc cx                   ;byte count of file name
         L3:
         inc bx
       jmp short L0

     L4:
     cmp dx,2
     jb >L5
     mov cs:[paralen],cl
     L5:
     pop bp
     ret

   cmdline endp


;--------------------------------------------------------------------
; CAPS procedure.  Capitalizes any string.
; Input: Near pointer to string, length of buffer
;--------------------------------------------------------------------

   caps proc near

     push bp
     mov bp,sp
     mov si,[bp+4]
     mov cx,[bp+6]
     L0:
       mov al,[si]
       cmp al,61h                      ;below a?
       jb >L1
       cmp al,7ah                      ;above z?
       ja >L1
         and al,0dfh                   ;clear the 6th bit
         mov [si],al
       L1:
         inc si
     loop L0
     pop bp
     ret

   caps endp


;--------------------------------------------------------------------
; CONVERT procedure.  Converts any Hex number to displayable
; decimal string.
; Input: Doubleword in dx:ax 
; Output: To screen automatically (from variable ascii_string)
;--------------------------------------------------------------------

   convert proc near

     mov si,10                         ;load base 10 into si
     mov bx,ax                         ;load low word into bx
     mov cx,dx                         ;load high word into cx
     mov dx,256                        ;set end-of-string flag
     push dx                           ;save it as last string byte
     L0:
       mov ax,cx                       ;load ax with high word
       xor dx,dx                       ;zero dx
       div si                          ;divide by 10, leave remainder in dx
       mov cx,ax                       ;save high word back in cx
       mov ax,bx                       ;load ax with low word
       div si                          ;divide by 10
       mov bx,ax                       ;save low word back in bx
       add dl,30h                      ;convert rem. to ascii
       push dx                         ;save as digit
     or bx,bx                          ;see if low word is zero
     jnz L0
     or cx,cx                          ;see if high word is zero
     jnz L0

     xor cx,cx
     mov di,offset ascii_string        ;point di to string

     L0:
       pop ax                          ;pop first digit
       cmp ax,256                      ;see if end-of-string
       je >L1
       stosb                           ;write it to string
       inc cx                          ;increase digit counter
     jmp short L0

     L1:
       cmp cx,1                        ;see if string is one digit long
       jne >L0                         ;if not, write string
       mov al,[di-1]                   ;move only digit to al
       mov ah,'0'                      ;move "0" into ah
       mov [di],al                     ;write digit as second letter
       mov [di-1],ah                   ;write "0" as first
       inc cx                          ;adjust counter to two

     L0:
       writefile 1,ascii_string,cx

     ret

   convert endp

;_TEXT ENDS
